/*
 * caller id related functions
 *
 * This file is part of ANT (Ant is Not a Telephone)
 *
 * Copyright 2002, 2003 Roland Stigge
 *
 */

#include "config.h"

/* regular GNU system includes */
#include <stdio.h>
#ifdef HAVE_STDLIB_H
  #include <stdlib.h>
#endif
#include <math.h>
#include <time.h>
#include <string.h>

/* GTK */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

/* own header files */
#include "callerid.h"
#include "session.h"
#include "util.h"
#include "isdn.h"

/* graphical symbols */
#include "in.xpm"
#include "out.xpm"

struct cid_row_t { /* _additional_ (invisible) data associated with a row */
  int marked_unanswered;
};

/*
 * returns a new data structure (for association with a row)
 * filled with default values
 */
static gpointer cid_row_new() {
  struct cid_row_t *rowdata =
    (struct cid_row_t *) malloc(sizeof(struct cid_row_t));
  
  rowdata->marked_unanswered = 0;
  return (gpointer) rowdata;
}

/*
 * callback to be called for additional row data on destruction of row
 */
static void cid_row_destroy(gpointer rowdata) {
  if (debug > 1)
    fprintf(stderr, "debug: destroying rowdata at %p.\n", rowdata);
  free(rowdata);
}

/* called when caller id monitor state check button is toggled */
void cid_toggle_cb(GtkWidget *widget _U_, gpointer data, guint action _U_) {
  struct session_t *session = (struct session_t *) data;

  if (GTK_CHECK_MENU_ITEM (session->cid_check_menu_item)->active) {
    gtk_widget_show(session->cid);
    session->option_show_callerid = 1;
  } else {
    gtk_widget_hide(session->cid);
    session->option_show_callerid = 0;
  }
}

/*
 * called on cid delete request
 *
 * -> deletes entry without confirmation
 *
 *  input:  widget["to_delete"]: number of cid_list item to delete
 *          data: session
 *  
 *  output: widget["to_delete"] entry will be removed
 */
static void cid_request_delete(GtkWidget *widget, gpointer data) {
  struct session_t *session = (struct session_t *) data;
  int row;

  if ((row = (int)gtk_object_get_data(GTK_OBJECT(widget), "to_delete"))) {
    cid_mark_row(session, row, 0);
    session->cid_num--;
    /* decrease before removal: select-row signal submission */
    gtk_clist_remove(GTK_CLIST(session->cid_list), row);
  
    gtk_object_remove_data(GTK_OBJECT(widget), "to_delete");
  }
}

/*
 * called on key pressed in cid list
 */
static gint cid_key_cb(GtkWidget *widget, GdkEventKey *event, gpointer data) {
  
  switch (event->keyval) {
    case GDK_Delete:     /* Delete dialog */
    case GDK_KP_Delete:
      /* portability: GtkCList.selection is private! */
      if (GTK_CLIST(((struct session_t *) data)->cid_list)->selection) {
        gtk_object_set_data(GTK_OBJECT(widget), "to_delete",
	    GTK_CLIST(((struct session_t *) data)->cid_list)->selection->data);
        cid_request_delete(widget, data);
      }
      return TRUE; /* event handled */
    default:
      return FALSE;
  }
}

/*
 * called on mouse button pressed in cid list
 */
static gint cid_mouse_cb(GtkWidget *widget _U_,
                         GdkEventButton *event, gpointer data) {
  struct session_t *session = (struct session_t *) data;
  
  if (event->button == 3) { /* popup menu */
    GtkMenu *menu = GTK_MENU(gtk_menu_new());
    GtkWidget *item = gtk_menu_item_new_with_label("Delete");
    gint row;
    gint column;

    if (gtk_clist_get_selection_info(GTK_CLIST(session->cid_list),
	                             event->x, event->y, &row, &column)) {

      gtk_object_set_data(GTK_OBJECT(item), "to_delete", (gpointer) row);
      gtk_signal_connect(GTK_OBJECT(item), "activate",
                         GTK_SIGNAL_FUNC(cid_request_delete),
		         data);
      gtk_widget_show(item);
      gtk_menu_append(menu, item);

      gtk_menu_popup(menu, NULL, NULL, NULL, NULL, event->button, event->time);
    }
    return TRUE;
  } else {
    return FALSE;
  }
}

/*
 * called on row select (to unmark lines)
 */
static void cid_select_row_cb(GtkCList *list _U_, gint row, gint column _U_,
                              GdkEventButton *button _U_, gpointer data) {
  struct session_t *session = (struct session_t *) data;

  if (row < session->cid_num)
    cid_mark_row(session, row, 0);
}

/*
 * Returns the Caller ID section as GtkWidget,
 *   sets CID related members in session
 * NOTE: caller has to gtk_widget_show() this widget
 */
GtkWidget *cid_new(struct session_t *session) {
  GtkWidget *frame;           /* frame with comment */
  GtkWidget *scrolled_window; /* the home for the clist */
  GtkWidget *clist;           /* the list itself */
  GtkStyle *style;            /* needed for preparing symbols */

  gchar *titles[] = {"Date/Time", "Type", "From", "To", "Duration" /*XXX, "Zone" */};

  frame = gtk_frame_new("Caller ID");
  gtk_container_set_border_width(GTK_CONTAINER(frame), 8);

  /* the scrolled window: no special adjustments */
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  /* always vertical scroll bar, horizontal... well... automatic means no ;) */
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_container_add(GTK_CONTAINER(frame), scrolled_window);
  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 8);
  gtk_widget_show(scrolled_window);

  /* the list widget */
  clist = gtk_clist_new_with_titles(sizeof(titles) / sizeof(gchar *), titles);
  
  /* default: GTK_SELECTION_MULTIPLE: */
  gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_BROWSE);

  gtk_clist_set_shadow_type(GTK_CLIST(clist), GTK_SHADOW_IN);
  gtk_clist_set_column_width(GTK_CLIST(clist), 0, 128);
  gtk_clist_set_column_width(GTK_CLIST(clist), 1, 30);
  gtk_clist_set_column_width(GTK_CLIST(clist), 2, 100);
  gtk_clist_set_column_width(GTK_CLIST(clist), 3, 100);
  gtk_clist_set_column_width(GTK_CLIST(clist), 4, 80);
  /*gtk_clist_set_column_width(GTK_CLIST(clist), 5, 80);*/
  gtk_container_add(GTK_CONTAINER(scrolled_window), clist);
  gtk_clist_column_titles_show(GTK_CLIST(clist));
  gtk_widget_set_usize(clist, 480, 128);

  gtk_signal_connect(GTK_OBJECT(clist), "key-press-event",
                     GTK_SIGNAL_FUNC(cid_key_cb), session);
  gtk_signal_connect(GTK_OBJECT(clist), "button-press-event",
                     GTK_SIGNAL_FUNC(cid_mouse_cb), session);
  gtk_signal_connect(GTK_OBJECT(clist), "select-row",
                     GTK_SIGNAL_FUNC(cid_select_row_cb), session);

  gtk_widget_show(clist);

  /* set up IN/OUT symbols */
  style = gtk_widget_get_style(session->main_window);
  session->symbol_in_pixmap = gdk_pixmap_create_from_xpm_d(
				 session->main_window->window,
				 &session->symbol_in_bitmap,
				 &style->bg[GTK_STATE_NORMAL],
				 (gchar **) in_xpm);
  session->symbol_out_pixmap = gdk_pixmap_create_from_xpm_d(
				 session->main_window->window,
				 &session->symbol_out_bitmap,
				 &style->bg[GTK_STATE_NORMAL],
				 (gchar **) out_xpm);
  session->cid = frame;
  session->cid_list = clist;
  session->cid_scrolled_window = scrolled_window;

  return frame;
}

/*
 * select last item in cid list and adjust view to end of list
 */
void cid_jump_to_end(struct session_t *session) {
  GtkAdjustment *adj;

  gtk_clist_select_row(GTK_CLIST(session->cid_list),
		       session->cid_num - 1, 0);

  adj = gtk_scrolled_window_get_vadjustment(
	  GTK_SCROLLED_WINDOW(session->cid_scrolled_window));
  gtk_adjustment_set_value(adj, fmax(adj->lower, adj->upper - adj->page_size));
}

/*
 * remove unneeded entries from caller id list (shorten log)
 */
void cid_normalize(struct session_t *session) {
  while (session->cid_num_max && session->cid_num > session->cid_num_max) {
    gtk_clist_remove(GTK_CLIST(session->cid_list), 0);
    session->cid_num--;
  }
}

/*
 * returns a dynamically allocated string in cid conformant time format
 * caller has got to take care of freeing this memory after usage
 */
static char *cid_timestring(time_t t) {
  char *date = (char *) malloc(20);
  int len;
  
  date[0] = '\1';
  len = strftime(date, 20, "%Y-%m-%d %H:%M:%S", localtime(&t));
  if (len == 0 && date[0] != '\0') {
    fprintf(stderr, "cid: Error calculating time with strftime.\n");
    return NULL;
  }

  return date;
}

/*
 * insert new entry at end of list
 *
 * from == NULL, if still unknown
 */
void cid_add_line(struct session_t *session,
		  enum call_type_t ct, gchar *from, gchar *to) {
  gchar *line[5];
  char *typestring;

  GdkPixmap *pixmap;
  GdkBitmap *bitmap;
 
  /* First: date */
  line[0] = cid_timestring(session->ring_time);
  line[1] = ""; /* call type */

  if (from == NULL) /* the from field */
    line[2] = "???";
  else
    line[2] = from;
  line[3] = to;

  line[4] = "???"; /* duration */

  /* create line */
  session->cid_num = gtk_clist_append(GTK_CLIST(session->cid_list), line) + 1;
  gtk_clist_set_row_data_full(GTK_CLIST(session->cid_list),
                              session->cid_num - 1,
			      cid_row_new(), cid_row_destroy);

  cid_normalize(session);

  /* actually set pixmap */
  if (ct == CALL_IN) {
    pixmap = session->symbol_in_pixmap;
    bitmap = session->symbol_in_bitmap;
    typestring = "IN";
  } else {
    pixmap = session->symbol_out_pixmap;
    bitmap = session->symbol_out_bitmap;
    typestring = "OUT";
  }
  gtk_clist_set_pixtext(GTK_CLIST(session->cid_list), session->cid_num - 1 , 1,
			typestring, 50, pixmap, bitmap);

  cid_jump_to_end(session);
  
  /* clean up */
  free(line[0]);
}

/*
 * set "date" field in last line
 */
void cid_set_date(struct session_t *session, time_t date) {
  if (date) {
    char *temp;
    gtk_clist_set_text(GTK_CLIST(session->cid_list),
		       session->cid_num - 1, 0, temp = cid_timestring(date));
    free(temp);
  }
}

/*
 * set "from" field in last line
 */
void cid_set_from(struct session_t *session, gchar *from) {
  if (from)
    gtk_clist_set_text(GTK_CLIST(session->cid_list),
		       session->cid_num - 1, 2, from);
}

/*
 * complete last line with "duration"
 *
 * if message == NULL, current time will be used to calculate a duration
 * if message != NULL, this string will be displayed in the Duration field (#4)
 */
void cid_set_duration(struct session_t *session, gchar *message) {
  char *buf = timediff_str(time(NULL), session->vcon_time);
  
  if (message)
    gtk_clist_set_text(GTK_CLIST(session->cid_list), session->cid_num - 1, 4,
		       message);
  else
    gtk_clist_set_text(GTK_CLIST(session->cid_list), session->cid_num - 1, 4,
		       buf);
  free(buf);
}

/*
 * Add a line to clist (e.g. from saved caller id file)
 */
void cid_add_saved_line(struct session_t *session, char *date, char *type,
			char *from, char *to, char *duration) {
  if (debug > 1)
    fprintf(stderr, "Caller ID add:\n"
	    "Date: |%s|, Type: |%s|, From: |%s|, To: |%s|, Dur: |%s|\n",
            date, type, from, to, duration);
  
  cid_add_line(session,
	       !strncmp(type, "IN", 2) ? CALL_IN : CALL_OUT, from, to);
  gtk_clist_set_text(GTK_CLIST(session->cid_list),
		     session->cid_num - 1, 0, date);
  gtk_clist_set_text(GTK_CLIST(session->cid_list),
		     session->cid_num - 1, 4, duration);
}

/*
 * Merges isdnlog data into existing cid list
 */
void cid_calls_merge(struct session_t *session) {
  int low, high, mid; /* bounds and helper index */
  time_t start = time(NULL)
                        - 60 * 60 * 24 * session->option_calls_merge_max_days;
  int callstime; /* time of calls file line */
  char *cidstr; /* time of cid_list entry */
  size_t linelength = 100; /* buffer size */ 
  char *line = (char *) malloc (linelength); /* buffer for getline() */
  FILE *f;

  char from[41]; /* data to read from calls file */
  char to[41];
  int duration;
  int date;
  char type[4];
  
  char *temp;
  int rel;
  int cidline; /* index into cid list */
  gchar *linearr[5]; /* line to insert to cid_list */
  GdkPixmap *pixmap;
  GdkBitmap *bitmap;

  char *cache; /* tracks successive dates */
  
  
  /* try to find isdnlog data file */
  if ((f = fopen(ISDN_ISDNLOG_FILE, "r"))) {
    /* do a binary search on the file for the desired starting time if needed */
    if (session->option_calls_merge_max_days) {
      low = 0;
      fseek(f, 0, SEEK_END);
      high = ftell(f);
      
      while (high - low > 200) {
        mid = (low + high) / 2;
        fseek(f, mid, SEEK_SET);
	getline(&line, &linelength, f); /* discard potentially partial line */
	if (-1 != getline(&line, &linelength, f)) {
	  if (!sscanf(line, "%*40[^|]|%*40[^|]|%*40[^|]|%*40[^|]|%*40[^|]|%d",
	              &callstime))
	    break;
	  if (callstime < start)
	    low = mid;
	  else
	    high = mid;
	} else
	  break;
      }
      fseek(f, low, SEEK_SET);
      getline(&line, &linelength, f);
    }
    
    /* read in all remaining lines: merge */
    cidline = 0;
    cache = strdup("");
    while (!feof(f)) {
      /* get new calls line */
      if (-1 == getline(&line, &linelength, f))	break;
      if (sscanf(line,
	         "%*40[^|]|%40[^|]|%40[^|]|%d|%*40[^|]|%d|%*40[^|]|%1c",
	         from, to, &duration, &date, type) != 5) {
	fprintf(stderr, "Warning: Incomplete data input from calls file.\n");
	break;
      }
      if ((temp = strchr(from, ' '))) *temp = '\0';
      if ((temp = strchr(to, ' '))) *temp = '\0';
      if (type[0] == 'I') {
	type[1] = 'N';
	type[2] = '\0';
        pixmap = session->symbol_in_pixmap;
        bitmap = session->symbol_in_bitmap;
      } else {
	type[1] = 'U';
	type[2] = 'T';
	type[3] = '\0';
        pixmap = session->symbol_out_pixmap;
        bitmap = session->symbol_out_bitmap;
      }
      
      /* prepare (text) line to merge */
      linearr[0] = cid_timestring((time_t)date);
      linearr[1] = type;
      if (from[0])
	linearr[2] = from;
      else
	linearr[2] = "???";
      if (to[0])
	linearr[3] = to;
      else
	linearr[3] = "???";
      if (duration < 0) {
	linearr[4] = strdup("(UNKNOWN)");
      } else {
	linearr[4] = timediff_str(duration, (time_t) 0);
      }

      /* where to merge it in? */
      rel = 1; /* init to merge by default */
      while (cidline < session->cid_num &&
        (gtk_clist_get_text(GTK_CLIST(session->cid_list), cidline, 0, &cidstr),
	 (rel = strcmp(cidstr, linearr[0])) < 0)) {
	
	cidline++;
      }

      if ((rel > 0 && strcmp(linearr[0], cache) > 0)
	  || cidline == session->cid_num) {
	/* only merge new and successive entries */
	
        gtk_clist_insert(GTK_CLIST(session->cid_list), cidline, linearr);
        gtk_clist_set_row_data_full(GTK_CLIST(session->cid_list), cidline,
                                    cid_row_new(), cid_row_destroy);
        gtk_clist_set_pixtext(GTK_CLIST(session->cid_list),
	    cidline , 1, type, 50, pixmap, bitmap);
	session->cid_num++;
      }
      free(cache);
      cache = linearr[0];
      free(linearr[4]);
    }

    free(linearr[0]);
    
    cid_normalize(session);
  
    if (fclose(f))
      fprintf(stderr, "Error closing " ISDN_ISDNLOG_FILE ".\n");
  } else { /* error on fopen() */
    fprintf(stderr, "Warning: Couldn't open " ISDN_ISDNLOG_FILE
	". Proceeding without it.\n");
  }
  free(line);
}

/*
 * - marks caller id row as unanswered
 * - sets window title to reflect number of unanswered calls
 *
 * input: session
 *        row     number of specified row
 *        state   1 for mark, 0 for unmark
 */
void cid_mark_row(struct session_t *session, int row, int state) {
  struct cid_row_t *rowdata = (struct cid_row_t *) gtk_clist_get_row_data(
                                            GTK_CLIST(session->cid_list), row);
  GdkColor color;
  char *title;

  if (rowdata->marked_unanswered != state) {
    if (state) {
      gdk_color_parse("#FF8888", &color);
      session->unanswered++;
      rowdata->marked_unanswered = 1;
    } else {
      gdk_color_parse("#FFFFFF", &color);
      session->unanswered--;
      rowdata->marked_unanswered = 0;
    }
    gtk_clist_set_background(GTK_CLIST(session->cid_list), row, &color);
    if (session->unanswered
	&& 0 < asprintf(&title, "ANT: %d unanswered", session->unanswered)) {
      gtk_window_set_title(GTK_WINDOW(session->main_window), title);
      free(title);
    } else {
      gtk_window_set_title(GTK_WINDOW(session->main_window), "ANT " VERSION);
    }
  }
}

